use std::sync::Mutex;

use chrono::{DateTime, Datelike, Timelike};
use chrono_tz::Europe::Berlin;

use esp_idf_hal::delay::Delay;
use esp_idf_sys::
    esp_restart
;
use log::error;
use once_cell::sync::Lazy;
use plant_hal::{CreatePlantHal, PlantCtrlBoard, PlantCtrlBoardInteraction, PlantHal};
use serde::{Deserialize, Serialize};

pub mod plant_hal;

#[derive(Serialize, Deserialize, Copy, Clone, Debug, PartialEq)]
enum OnlineMode {
    Offline,
    Wifi,
    SnTp,
    Online,
}

pub static BOARD_ACCESS: Lazy<Mutex<PlantCtrlBoard>> = Lazy::new(|| PlantHal::create().unwrap());


fn safe_main() -> anyhow::Result<()> {
    // It is necessary to call this function once. Otherwise some patches to the runtime
    // implemented by esp-idf-sys might not link properly. See https://github.com/esp-rs/esp-idf-template/issues/71
    esp_idf_svc::sys::link_patches();

    // Bind the log crate to the ESP Logging facilities
    esp_idf_svc::log::EspLogger::initialize_default();

    if esp_idf_sys::CONFIG_MAIN_TASK_STACK_SIZE < 25000 {
        error!(
            "stack too small: {} bail!",
            esp_idf_sys::CONFIG_MAIN_TASK_STACK_SIZE
        );
        return Ok(());
    }

    log::info!("Startup Rust");

    println!("Board hal init");
    let mut board: std::sync::MutexGuard<'_, PlantCtrlBoard<'_>> = BOARD_ACCESS.lock().unwrap();

    let time = board.time();
    let mut cur = match time {
        Ok(cur) => cur,
        Err(err) => {
            log::error!("time error {}", err);
            DateTime::from_timestamp_millis(0).unwrap()
        }
    };
    //check if we know the time current > 2020
    if cur.year() < 2020 {
            //assume TZ safe times ;)
            cur = *cur.with_hour(15).get_or_insert(cur);
        
    }

    println!("cur is {}", cur);


    let mut online_mode: OnlineMode;
 
    println!("attempting to connect wifi");
    let mut ssid: heapless::String<32> = heapless::String::new();
    let mut password: heapless::String<64> = heapless::String::new();

    ssid.push_str("C3MA").unwrap();
    password.push_str("penis").unwrap();

    let ip_address: Option<String>;
    match board.wifi(ssid, Option::Some(password), 10000) {
        Ok(ip_info) => {
            ip_address = Some(ip_info.ip.to_string());
            online_mode = OnlineMode::Wifi;
        }
        Err(_) => {
            panic!("wifi down");
        }
    }

    if online_mode == OnlineMode::Wifi {
        match board.sntp(1000 * 5) {
            Ok(new_time) => {
                cur = new_time;
                online_mode = OnlineMode::SnTp;
            }
            Err(err) => {
                println!("Sntp error {}", err);
                online_mode = OnlineMode::SnTp;
            }
        }
    }

    println!("Running logic at utc {}", cur);
    let europe_time = cur.with_timezone(&Berlin);
    println!("Running logic at europe/berlin {}", europe_time);

    //do mqtt before config check, as mqtt might configure
    if online_mode == OnlineMode::SnTp {
        match board.mqtt() {
            Ok(_) => {
                println!("Mqtt connection ready");
                online_mode = OnlineMode::Online;
            }
            Err(err) => {
                panic!("Could not connect mqtt due to {}", err);
                            }
        }
    }

    if online_mode == OnlineMode::Online {
        match ip_address {
            Some(add_some) => {
                let _ = board.mqtt_publish("/firmware/address", add_some.as_bytes());
            }
            None => {
                let _ = board.mqtt_publish("/firmware/address", "N/A?".as_bytes());
            }
        }
        let _ = board.mqtt_publish(
            "/firmware/last_online",
            europe_time.to_rfc3339().as_bytes(),
        );
        let _ = board.mqtt_publish("/state", "online".as_bytes());

    }
    drop(board);

    loop {
        //wait till earth ends
        Delay::new(1000).delay_ms(1000);
        let mut lock = BOARD_ACCESS.lock().unwrap();
        let position = lock.get_position();
        lock.mqtt_publish("/position", format!("{}", position).as_bytes()).unwrap();
    }
}

fn main() {
    let result = safe_main();
    match result {
        Ok(_) => {
            println!("Main app finished, restarting");
            unsafe { esp_restart() };
        }
        Err(err) => {
            panic!("Failed main {}", err);
        }
    }
}
//error codes
//error_reading_config_after_upgrade
//error_no_config_after_upgrade
//error_tank_sensor_fault
